<?php

namespace InSegment\ApiCore\Scopes;

use Illuminate\Support\Facades\DB;

use InSegment\ApiCore\Services\Transactor;

use Illuminate\Database\Query\Builder as QueryBuiler;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Database\Eloquent\Model;
use Illuminate\Database\Eloquent\Scope;

/**
 * A global scope to ease the use of Transactor with Eloquent by the help of UNIONs
 */
class TransactionSearchScope implements Scope
{
    /**
     * Properties of the query Bilder which will be shared between UNIONS
     * 
     * @var type 
     */
    private static $builderPropertiesToAssociate = [
        'connection', 'grammar', 'processor', 'bindings', 'aggregate', 'wheres'
    ];

    /**
     * A set of classes having this scope added
     * 
     * @var array
     */
    private static $applied = [];
    
    /**
     * The indication of scope is disabled
     * 
     * @var bool 
     */
    private static $disabled = false;
    
    /**
     * Adds the global scope to the class, if it wasn't already added there
     * 
     * @return null
     */
    public static function applyTo(string $class)
    {
        if (!self::isApplied($class) && Transactor::doesHaveTables($class)) {
            $class::addGlobalScope(new static($class));
            self::$applied[$class] = true;
        }
    }
    
    /**
     * Checks if the scope is applied to the class
     * 
     * @param string $class
     * @return bool
     */
    public static function isApplied(string $class): bool
    {
        return isset(self::$applied[$class]);
    }
    
    /**
     * Disable the scope
     */
    public static function disable()
    {
        static::$disabled = true;
    }
    
    /**
     * Class of the Eloquent Model this scope works on
     * 
     * @var string 
     */
    private $applyClass;
    
    /**
     * Private to prevent creation from outside
     * Sets applyClass
     * 
     * @param string $class
     */
    private function __construct(string $class) {
        $this->applyClass = $class;
    }
    
	/**
	 * Adds UNION on the three tables: the main table, the update table and the insert table
     * No modifications possible while under this global scope.
     * I.e, you will need $query->withoutGlobalScope(TransactionSearchScope::class) to be able to UPDATE or INSERT
     * 
     * UPDATE table takes priority over the main table and INSERT table.
     * INSERT table takes priority over the main table
     * Main table is the last resort
	 * 
	 * @param \Illuminate\Database\Eloquent\Builder $builder
	 * @param \Illuminate\Database\Eloquent\Model $model
	 * @return null
	 */
	public function apply(Builder $builder, Model $model)
	{
        if (static::$disabled) {
            return;
        }
        
        $table = $model->getTable();
        $builder->select("{$table}.*")->groupBy("{$table}.{$model->getKeyName()}");
        
        $baseQuery = $builder->getQuery();

        $updateTable = Transactor::getUpdateTable($this->applyClass);
        $updateQuery = $this->modifyQuery(DB::table("{$updateTable} as {$table}"), $baseQuery);

        $insertTable = Transactor::getInsertTable($this->applyClass);
        $insertQuery = $this->modifyQuery(DB::table("{$insertTable} as {$table}"), $baseQuery);

        $mainQuery = $this->modifyQuery(DB::table($table), $baseQuery);

        $updateQuery
            ->union($insertQuery)
            ->union($mainQuery);

        // Select from sumquery with unions, but LIE AND PRETEND IT'S THE SOURCE TABLE
        $baseQuery->from = DB::raw("({$updateQuery->toSql()}) as `{$table}`");
	}
    
    /**
     * Hack several builders into sharing conditions and bindings
     * 
     * @param \Illuminate\Database\Query\Builder $target
     * @param \Illuminate\Database\Query\Builder $source
     * @return \Illuminate\Database\Query\Builder
     */
    private function modifyQuery(QueryBuiler $target, QueryBuiler $source)
    {
        foreach (self::$builderPropertiesToAssociate as $property) {
            $target->$property = &$source->$property;
        }
        
        return $target;
    }
    
}